/* Purpose:
Get familiar with the eCos exception handling.
Note that the eCos kernel must be configured for handling exceptions.
*/
#include <cyg/kernel/kapi.h>     // All the kernel specific stuff
#include <stdio.h>               // printf()
#include <stdlib.h>              // exit()
#include <cyg/hal/hal_intr.h>    // interrupt and exception vectors

#define NTHREADS 1
#define STACKSIZE 5000 //4096

// thread function prototype
void thread_prog(cyg_addrword_t data);

/* now declare (and allocate space for) some kernel objects */
static cyg_thread thread_obj[NTHREADS];	/* space for thread objects */
static cyg_handle_t thread[NTHREADS];   /* handles for the threads */
static char stack[NTHREADS][STACKSIZE];	/* space for n threads stacks, n*4K */


// New System Call exception handler.
void my_exception_handler(cyg_addrword_t data, cyg_code_t number,
			  cyg_addrword_t info)
{
  printf(" --> Exception: ");
  switch(number) {
  case CYGNUM_HAL_EXCEPTION_ILLEGAL_INSTRUCTION:
    printf("Illegal instruction\n");
    break;
  case CYGNUM_HAL_EXCEPTION_PRIVILEGED_INSTRUCTION:
    printf("Priveleged instruction\n");
    break;
  case CYGNUM_HAL_EXCEPTION_FPU_NOT_AVAIL:
    printf("No FPU available\n");
    break;
  default:
    printf("number %d\n", number);
    break;
  }

  //exit(0);
  //  exit(1); // if eCos assertions are active, this triggers them
}


// This is the application entry point
void cyg_user_start(void)
{
  printf("Entering cyg_user_start() function\n");

  cyg_thread_create(4, thread_prog, (cyg_addrword_t) 0,
		    "Thread A", (void *) stack[0], STACKSIZE,
		    &thread[0], &thread_obj[0]);
  cyg_thread_resume(thread[0]);
}



/* this is a simple program which runs in a thread */
void thread_prog(cyg_addrword_t data)
{
  int message = (int) data;
  int delay;
  volatile float a, b, c;
  int i;

  printf("Beginning execution; thread data is %d\n", message);

  // register the the exception handler
/*  cyg_exception_set_handler(CYGNUM_HAL_EXCEPTION_PRIVILEGED_INSTRUCTION,
			    my_exception_handler, 0, NULL, NULL);
  cyg_exception_set_handler(CYGNUM_HAL_EXCEPTION_FPU_NOT_AVAIL,
			    my_exception_handler, 0, NULL, NULL);
  cyg_exception_set_handler(CYGNUM_HAL_EXCEPTION_OTHERS,
			    my_exception_handler, 0, NULL, NULL);
  cyg_exception_set_handler(CYGNUM_HAL_EXCEPTION_ILLEGAL_INSTRUCTION,
			    my_exception_handler, 0, NULL, NULL);
*/
  for (i=16; i<27; i++) {		// all exceptions
    cyg_exception_set_handler(i, my_exception_handler, 0, NULL, NULL);
  }

  cyg_thread_delay(100);
  a = 2.0;
  b = 0.0;
  c = a/b; // no exception!?
  printf("Entering infinite loop\n");

  for (;;) {
    c = a/b; // no exception!?
    delay = 100;
    printf("Thread %d: and now a delay of %d clock ticks\n", message, delay);
    cyg_thread_delay(delay);
    //cyg_exception_call_handler(cyg_thread_self(),10,0);

    // mem_address_not_aligned trap, tt=0x07
/*    asm("sethi %hi(0x80001000), %l0");
    asm("or %l0, %lo(0x80001003), %l0");
    asm("st %l1, [%l0]");
*/

    // data_store_error, tt=0x2b
/*    asm("sethi %lo(3), %l0");
    asm("st %l1, [%l0]");
*/

    //asm("ta 16; nop");
//    asm("ta 18; nop");
    asm("rd %psr, %l0");
    asm("wr %l0, %psr"); // generates ILLEGAL INSTRUCTION exception vec=18, tt=2
    
    //asm("mov %%psr, %%l0":::"%l0"); // legal
    //asm("mov %psr, %g0"); // legal
    //asm("cpop1 1, %c0, %c1, %c2"); // coprocessor register
    //asm("mov %l0, %c2"); // coprocessor register
  }
}

