/*=============================================================================
//
//	vectors.S
//
//	SPARC vectors and bootup code
//
//=============================================================================
//####ECOSGPLCOPYRIGHTBEGIN####
// -------------------------------------------
// This file is part of eCos, the Embedded Configurable Operating System.
// Copyright (C) 1998, 1999, 2000, 2001, 2002 Red Hat, Inc.
//
// eCos is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later version.
//
// eCos is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with eCos; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
//
// As a special exception, if other files instantiate templates or use macros
// or inline functions from this file, or you compile this file and link it
// with other works to produce a work based on this file, this file does not
// by itself cause the resulting work to be covered by the GNU General Public
// License. However the source code for this file must still be made available
// in accordance with section (3) of the GNU General Public License.
//
// This exception does not invalidate any other reasons why a work based on
// this file might be covered by the GNU General Public License.
//
// Alternative licenses for eCos may be arranged by contacting Red Hat, Inc.
// at http://sources.redhat.com/ecos/ecos-license/
// -------------------------------------------
//####ECOSGPLCOPYRIGHTEND####
//=============================================================================
//#####DESCRIPTIONBEGIN####
//
// Author(s): 	hmt
// Contributors:hmt,
//              Gaisler Research (Konrad Eisele<eiselekd@web.de>) (fpu support)
// Date:	1998-12-15
// Purpose:	SPARC vector code
// Description:	This file contains the code which hangs off SPARC vectors
//              including reset; it handles register under/overflow as well
//              as bootup, anything else is deferred to the default interrupt
//              or exception vsrs respectively.  See vec_[ix]vsr.S ...
//
//####DESCRIPTIONEND####
//
//===========================================================================*/

!-----------------------------------------------------------------------------
	
//	.file	"vectors.S"

!-----------------------------------------------------------------------------

			
#include <pkgconf/system.h>	
#include <pkgconf/hal.h>

#include <pkgconf/hal_sparc.h>
#include CYGBLD_HAL_PLATFORM_H		// Platform config file

#ifdef CYGPKG_KERNEL
# include <pkgconf/kernel.h>
#else
# undef CYGIMP_HAL_COMMON_INTERRUPTS_USE_INTERRUPT_STACK
# undef CYGFUN_HAL_COMMON_KERNEL_SUPPORT
#endif
			
#include <cyg/hal/arch.inc>
				
#if defined( CYGPKG_HAL_SPARCLITE_SIM ) || \
    defined( CYGIMP_HAL_COMMON_INTERRUPTS_USE_INTERRUPT_STACK ) || \
    defined( CYGPKG_HAL_SPARC_LEON ) || \
    defined( CYGPKG_HAL_SPARC_LEON3 ) || \
    defined( CYGPKG_HAL_SPARC_ERC32 )
#define BOOTUPSTACK_IS_INTERRUPTSTACK
#endif
	
//#define CYGHWR_HAL_SPARC_MULTIPLE_VECTOR_TRAPPING

#ifndef CYGHWR_HAL_SPARC_HAS_ASR17
#ifndef CYGHWR_HAL_SPARC_MULTIPLE_VECTOR_TRAPPING
#error Single Vector Trapping (SVT) demands ASR17 
#endif
#endif
	
#if !defined(CYGHWR_HAL_SPARC_MULTIPLE_VECTOR_TRAPPING) && !defined(CYGPKG_HAL_SPARC_LEON3)
#error Single Vector Trapping (SVT) demands Leon3 model 
#endif
            	
#ifdef CYG_HAL_STARTUP_ROM
# ifndef CYGHWR_HAL_SPARC_MULTIPLE_VECTOR_TRAPPING

! ROM startup and Single Vector Trapping demands a copy to RAM.
! Otherwise it may be configured in, but it is not required.
	
#  ifndef CYGIMP_HAL_SPARC_COPY_VECTORS_TO_RAM
#   define CYGIMP_HAL_SPARC_COPY_VECTORS_TO_RAM
#  endif

# endif
#endif

!------------------------------------------------------------------------

#include <cyg/hal/vectors.h>

#define DELAYS_AFTER_WRPSR_SAME_WINDOW
#define DELAYS_AFTER_WRWIM

!------------------------------------------------------------------------

#ifdef CYGHWR_HAL_SPARC_MULTIPLE_VECTOR_TRAPPING

	.macro	VECTOR_SYSCALL _vector _handler
	.p2align 4
         mov   %psr, %l0 
         sethi %hi(\_handler), %l4 
         jmp   %l4+%lo(\_handler)
         mov   \_vector, %l3
        .endm

	.macro	VECTOR_DUMMY
	.p2align 4
	jmp	%l1
	rett	%l2
	nop
	nop
        .endm
	
	.macro	VECTOR_EXCEPTION
	.p2align 4
	rd	%tbr, %l3
	rd	%psr, %l0
	ba	__entry_exception
	 and	%l3, TT_MASK, %l4
	.endm
	
	.macro	VECTOR_INTERRUPT level
	.p2align 4
	rd	%psr, %l0
	mov	\level << 2, %l5
	ba	__entry_interrupt
	 mov	\level << 4, %l4
	.endm
	
	.macro	VECTOR_CODE_WIM name
	.p2align 4
	ba	__entry_\name
	 rd	%wim, %l0
	.endm
	
	.macro	VECTOR_FPDIS 
	.p2align 4
	ba	__entry_fpdis
 	 rd	%psr, %l0
	.endm
		
#endif // CYGHWR_HAL_SPARC_MULTIPLE_VECTOR_TRAPPING

!------------------------------------------------------------------------
! The start of the code; this is the entry point:
	
	.section ".vectors","ax"

	.global rom_vectors
rom_vectors:
trap_table:             
#ifdef CYGHWR_HAL_SPARC_MULTIPLE_VECTOR_TRAPPING
	! this code goes to the real reset handler, it will be
	! overwritten by the start of vectoring handler...
	b	genuine_reset
	 nop
	! usually drop through to:
	ba	__entry_exception	! reset becomes an exception
	 and	%l3, TT_MASK, %l4	! once we are running
	
	VECTOR_EXCEPTION	 	!   1
	VECTOR_EXCEPTION	 	!   2
	VECTOR_EXCEPTION	 	!   3
	VECTOR_FPDIS			!   4 fp disabled
	
	VECTOR_CODE_WIM wover	 	!   5 window overflow
	VECTOR_CODE_WIM wunder		!   6 window underflow
	VECTOR_EXCEPTION		!   7
	VECTOR_EXCEPTION		!   8
	VECTOR_EXCEPTION		!   9

	VECTOR_EXCEPTION		!  10
	VECTOR_EXCEPTION		!  11
	VECTOR_EXCEPTION		!  12
	VECTOR_EXCEPTION		!  13
	VECTOR_EXCEPTION		!  14
	
	VECTOR_EXCEPTION		!  15
	VECTOR_EXCEPTION		!  16
	VECTOR_INTERRUPT  1		!  17 interrupt_level_1
	VECTOR_INTERRUPT  2		!  18 interrupt_level_2
	VECTOR_INTERRUPT  3		!  19 interrupt_level_3
	
	VECTOR_INTERRUPT  4		!  20 interrupt_level_4
	VECTOR_INTERRUPT  5		!  21 interrupt_level_5
	VECTOR_INTERRUPT  6		!  22 interrupt_level_6
	VECTOR_INTERRUPT  7		!  23 interrupt_level_7
	VECTOR_INTERRUPT  8		!  24 interrupt_level_8
	
	VECTOR_INTERRUPT  9		!  25 interrupt_level_9
	VECTOR_INTERRUPT 10		!  26 interrupt_level_10
	VECTOR_INTERRUPT 11		!  27 interrupt_level_11
	VECTOR_INTERRUPT 12		!  28 interrupt_level_12
	VECTOR_INTERRUPT 13		!  29 interrupt_level_13
	
	VECTOR_INTERRUPT 14		!  30 interrupt_level_14
	VECTOR_INTERRUPT 15		!  31 interrupt_level_15
	VECTOR_EXCEPTION		!  32
	VECTOR_EXCEPTION		!  33
	VECTOR_EXCEPTION		!  34

	VECTOR_EXCEPTION		!  35
	VECTOR_EXCEPTION		!  36 cp disabled
	VECTOR_EXCEPTION		!  37
	VECTOR_EXCEPTION		!  38
	VECTOR_EXCEPTION		!  39
	
	.rept 88			! 40-127 is 88 of them
	VECTOR_EXCEPTION		! whichever
	.endr

        VECTOR_SYSCALL 0x80 syscall     ! 128 syscall
        
	.rept 127			! 129-255 is 127 of them
	VECTOR_EXCEPTION		! whichever
	.endr

	
#endif // CYGHWR_HAL_SPARC_MULTIPLE_VECTOR_TRAPPING

//_______________________________________________
#ifndef CYGHWR_HAL_SPARC_MULTIPLE_VECTOR_TRAPPING
real_vector:
	rd	%tbr, %l3
real_vector_reset: 
	rd	%psr, %l0
	! here,locals have been set up as follows:
	! %l0 = psr
	! %l1 = pc	
	! %l2 = npc
	! %l3 = tbr
	and	%l3, TT_IS_INTR_MASK, %l4
	cmp	%l4, TT_IS_INTR_VALUE
	bne	not_an_interrupt	! delay slot does not matter
	! here be the pre-vector interrupt handler
interrupt:
	and	%l3, 0x0f0, %l4		! get an interrupt number out
	srl	%l4, 2, %l5		! to a word address offset
#endif // !CYGHWR_HAL_SPARC_MULTIPLE_VECTOR_TRAPPING
//''''''''''''''''''''''''''''''''''''''''''''''''''
        
__entry_interrupt:
#ifdef CYGPKG_HAL_HAS_TMU	// TMU Time Management Unit
	// if tmu irq; jump to exception vector service routine instead
	sethi %hi(CYGPKG_HAL_TMU_IRQ), %l6
	or %l6, %lo(CYGPKG_HAL_TMU_IRQ), %l6
	//or %g0, %lo(CYGPKG_HAL_TMU_IRQ), %l6 // save one instruction (g0 should be zero)
	srl	%l4, 4, %l3			// vector number into %l3
	cmp %l6, %l3 			// branch if tmu irq==current irq
	beq __entry_exception
	// set the active TMU timer
	sethi	%hi(CYGPKG_TMUADDR), %l6
	or	%l6, %lo(CYGPKG_TMUADDR + 4), %l6	// tsel addr
	st %l3, [ %l6 ]			// store vector num in timer select reg
#endif //  CYGPKG_HAL_HAS_TMU
	sethi	%hi(hal_vsr_table), %l6
	or	%l6, %lo(hal_vsr_table), %l6
	ld	[ %l6 + %l5 ], %l6	! get vector in %l6
	jmp	%l6			! and go there
	srl	%l4, 4, %l3		! vector number into %l3: so that
					! interrupts and exceptions/traps
					! have the same API to VSRs

//_______________________________________________
#ifndef CYGHWR_HAL_SPARC_MULTIPLE_VECTOR_TRAPPING
not_an_interrupt:
	and	%l3, TT_MASK, %l4
	cmp	%l4, (TRAP_WUNDER << TT_SHL)

	bne,a	not_window_underflow
	cmp	%l4, (TRAP_WOVER << TT_SHL)	! (if taken)

	! here be the window underflow handler:
window_underflow:
	! CWP is trap handler
	! CWP + 1 is trapped RESTORE instruction
	! CWP + 2 is invalid context which must be restored
	! CWP + 3 is next invalid context

	wr	%l0, %psr		! restore the condition flags
					! (CWP is unchanged)
	! the following instructions delay enough; no need for NOPs
	rd	%wim, %l0		! get the wim
#endif // !CYGHWR_HAL_SPARC_MULTIPLE_VECTOR_TRAPPING
//''''''''''''''''''''''''''''''''''''''''''''''''''
        
__entry_wunder:	
	sll	%l0, 1, %l3		! Rotate wim left
	srl	%l0, __WINSIZE-1, %l0
#if __WINSIZE > 3        
	wr	%l0, %l3, %wim		! Install the new wim
#else
	wr	%g0, %g0, %wim		! we need 2 windows for save,save
#endif

#ifdef DELAYS_AFTER_WRWIM
	nop				! are these delays needed?
	nop				! (following restore uses wim)
	nop
#endif
	restore				! Users window
	restore				! Her callers window (now valid)
	
	ldd	[%sp + 0 * 4], %l0	! restore L & I registers
	ldd	[%sp + 2 * 4], %l2
	ldd	[%sp + 4 * 4], %l4
	ldd	[%sp + 6 * 4], %l6

	ldd	[%sp + 8 * 4], %i0
	ldd	[%sp + 10 * 4], %i2
	ldd	[%sp + 12 * 4], %i4
	ldd	[%sp + 14 * 4], %i6

	save				! Back to trap window
	save	

#if __WINSIZE <= 3        
	wr	%l0, %l3, %wim		! Install the new wim (current)
#ifdef DELAYS_AFTER_WRWIM
	nop				! are these delays needed?
	nop				! 
	nop
#endif
#endif
        
	jmp	%l1
	rett	%l2

//_______________________________________________
#ifndef CYGHWR_HAL_SPARC_MULTIPLE_VECTOR_TRAPPING
        .global not_window_underflow
not_window_underflow:
	bne,a	not_window_overflow
	cmp	%l4, (TRAP_FPDIS << TT_SHL)	! (if taken)

	! here be the window overflow handler:
window_overflow:
	! CWP + 1 is caller whose SAVE bounced
	! CWP is trap handler = invalid window
	! CWP - 1 is next invalid window which needs to be saved

	wr	%l0, %psr		! restore the condition flags
        				! (CWP is unchanged)
	! the following instructions delay enough; no need for NOPs
	rd	%wim, %l0
#endif // !CYGHWR_HAL_SPARC_MULTIPLE_VECTOR_TRAPPING
//''''''''''''''''''''''''''''''''''''''''''''''''''

__entry_wover:	
	mov	%g1, %l3		! Save g1, we use it to hold the wim
	srl	%l0, 1, %g1		! Rotate wim right
	sll	%l0, __WINSIZE-1, %l0
	or	%l0, %g1, %g1

	save				! Slip into next window
	mov	%g1, %wim		! Install the new wim
					! (invalidates current window!)

	std	%l0, [%sp + 0 * 4]	! save L & I registers
	std	%l2, [%sp + 2 * 4]
	std	%l4, [%sp + 4 * 4]
	std	%l6, [%sp + 6 * 4]

	std	%i0, [%sp + 8 * 4]
	std	%i2, [%sp + 10 * 4]
	std	%i4, [%sp + 12 * 4]
	std	%i6, [%sp + 14 * 4]

	restore				! Go back to trap window.
	mov	%l3, %g1		! Restore %g1

	jmpl	%l1,  %g0
	rett	%l2

//_______________________________________________
#ifndef CYGHWR_HAL_SPARC_MULTIPLE_VECTOR_TRAPPING
        .global not_window_overflow
not_window_overflow:
	bne,a	not_fpdis
	cmp	%l4, (0x80 << TT_SHL)	! (if taken)

	! here be the entryfpdis handler:
entryfpdis:
	wr	%l0, %psr
	nop
	nop
	nop
	ba __entry_fpdis
	 nop
#endif // !CYGHWR_HAL_SPARC_MULTIPLE_VECTOR_TRAPPING
//''''''''''''''''''''''''''''''''''''''''''''''''''

#ifndef CYGPKG_HAL_SMP_CPU_MAX
#define CYGPKG_HAL_SMP_CPU_MAX 1
#endif 
				
	.data
	.global cyg_hal_fpustate_owner
cyg_hal_fpustate_owner:
	.long	0				! pointer to FPU owning context
	.global cyg_hal_fpustate_current	
cyg_hal_fpustate_current:	
	.long	0				! pointer to current threads FPU context

	.section ".vectors","ax"
	
	.global __entry_fpdis
__entry_fpdis:

#ifndef CYGHWR_HAL_SPARC_FPU
	ta 0
	nop
#else	
#ifndef CYGPKG_HAL_SPARC_LAZY_FPU_SWITCH
	set 0x1000,%l3
	or %l0,%l3,%l0	
	wr %l3,0, %psr		! restore the condition flags, enable fpu
	nop
	nop
	nop
	jmpl	%l1,  %g0
	rett	%l2			
#else	
	set 0x1000,%l3
	or %l0,%l3,%l0	
	or %l0,0xf00, %l3		! PIL up to 15, enable fpu
	wr %l3,0, %psr		! restore the condition flags, enable fpu
	nop
	nop
	nop
		
	set     cyg_hal_fpustate_current,%l4
	ld [%l4],%l4
	set     cyg_hal_fpustate_owner,%l5
	ld [%l5],%l5
	cmp	%l4,%l5
	beq	mpfnostore
	 nop
	cmp    %g0,%l5
	 beq	mpfstore
	nop
	
	std	%f0, [%l5 + 0x00]
	std	%f2, [%l5 + 0x08]
	std	%f4, [%l5 + 0x10]
	std	%f6, [%l5 + 0x18]
	std	%f8, [%l5 + 0x20]
	std	%f10, [%l5 + 0x28]
	std	%f12, [%l5 + 0x30]
	std	%f14, [%l5 + 0x38]
	std	%f16, [%l5 + 0x40]
	std	%f18, [%l5 + 0x48]
	std	%f20, [%l5 + 0x50]
	std	%f22, [%l5 + 0x58]
	std	%f24, [%l5 + 0x60]
	std	%f26, [%l5 + 0x68]
	std	%f28, [%l5 + 0x70]
	std	%f30, [%l5 + 0x78]
	
	st	%fsr, [%l5 + 0x80]
	
mpfstore:	
	set     cyg_hal_fpustate_owner,%l6
	st %l4,[%l6]
	cmp    %g0,%l4
	 beq	mpfnostore
	nop
	
	ldd	[%l4 + 0x00], %f0
	ldd	[%l4 + 0x08], %f2
	ldd	[%l4 + 0x10], %f4
	ldd	[%l4 + 0x18], %f6
	ldd	[%l4 + 0x20], %f8
	ldd	[%l4 + 0x28], %f10
	ldd	[%l4 + 0x30], %f12
	ldd	[%l4 + 0x38], %f14
	ldd	[%l4 + 0x40], %f16
	ldd	[%l4 + 0x48], %f18
	ldd	[%l4 + 0x50], %f20
	ldd	[%l4 + 0x58], %f22
	ldd	[%l4 + 0x60], %f24
	ldd	[%l4 + 0x68], %f26
	ldd	[%l4 + 0x70], %f28
	ldd	[%l4 + 0x78], %f30
	
	!ld	[%l4 + 0x80], %fsr
#ifdef GRFPU_NON_STD		
	ld      [%l4 + 0x80], %l5
	set	0x00400000, %l6
	or	%l5, %l6, %l5
	st	%l5, [%l4 + 0x80]
#endif	
	ld	[%l4 + 0x80], %fsr					

mpfnostore:	
	wr	%l0,0, %psr		! restore the condition flags, enable fpu
	nop
	nop
	nop
	
	jmpl	%l1,  %g0
	rett	%l2
#endif
#endif

//_______________________________________________
#ifndef CYGHWR_HAL_SPARC_MULTIPLE_VECTOR_TRAPPING
        .global not_fpdis
not_fpdis:      
	beq,a   syscall
         srl   %l4,TT_SHL, %l3           ! (if taken)
        
        cmp	%l4, (0x0 << TT_SHL)	
        beq	genuine_reset
	
        srl	%l4, 4, %l4		
#endif // CYGHWR_HAL_SPARC_MULTIPLE_VECTOR_TRAPPING
//''''''''''''''''''''''''''''''''''''''''''''''''''
        
#ifdef CYGHWR_HAL_SPARC_MULTIPLE_VECTOR_TRAPPING
	// ADDITIONAL code to provide an entry point:
__entry_exception:
	srl	%l4, 4, %l4
#endif // CYGHWR_HAL_SPARC_MULTIPLE_VECTOR_TRAPPING
not_exception:      
	! from here on in, %l4 is the trap number in clear
	cmp	%l4, 128 
	bge	1f
	mov	0, %l5			! offset 0 for user traps

	cmp	%l4, 36			! coprocessor special case
	beq	1f
	mov	4, %l5			! ...treated as FP, code 4

	cmp	%l4, 10
	bge	1f
	mov	10, %l5			! offset 10 for "others"

	! if we are here, the trap number is 1-9 inclusive
	! so put it in %l5 and drop through...
	mov	%l4, %l5
1:	
	or	%l5, 16, %l5		! offset into table is 16... for traps.
	sll	%l5,  2, %l5		! to a word address offset
	sethi	%hi(hal_vsr_table), %l6
	or	%l6, %lo(hal_vsr_table), %l6
	ld	[ %l6 + %l5 ], %l6	! get vector in %l6
	jmp	%l6			! and go there
	srl	%l5, 2, %l3		! vector number into %l3: so that
					! interrupts and exceptions/traps
					! have the same API to VSRs
	! NB that~s eCos vector number not TRAP number above.
	
	! and that is the end of the pre-vector trap handler	

	.global rom_vectors_end, real_vector_instructions
rom_vectors_end:
		
	! these instructions are copied into the reset vector
	! after startup to _not_ branch to the genuine_reset code below
real_vector_instructions:
	rd	%tbr, %l3
	rd	%psr, %l0


	.global	reset_vector
reset_vector:

  ! .global start
  .global start_ecos
  /* ------- */
  .weak	start
  .set	start,start_ecos
  	
start_ecos:
#ifndef CYGHWR_HAL_SPARC_MULTIPLE_VECTOR_TRAPPING
        mov %g0,%l3
        b real_vector_reset
         nop
#else
        b trap_table
         nop
#endif

	! genuine reset code called from time zero:
genuine_reset:				! set psr, mask interrupts & traps
	sub	%sp, 24 * 4, %fp	! fresh frame including
					! arg spill area for callees
	mov     %fp,%g5                 ! save sp,fp
	mov     %sp,%g6  
	sethi	%hi(trap_table), %g1
	andn	%g1, 0xfff, %g1		! should not be needed
	wr	%g1, %tbr		! Traps are at trap_table
#ifdef CYGPKG_HAL_SPARC_LAZY_FPU_SWITCH		
	set	0x01fc0 + __WIN_INIT, %g1
#else
	set	0x0fc0 + __WIN_INIT, %g1
#endif	
	wr	%g0, %g1, %psr		! mode = prevMode = S, CWP=7	
	!wr	%g0, 0xfc0 + __WIN_INIT, %psr ! mode = prevMode = S, CWP=7
	wr	%g0, 0, %wim		! No invalid windows (yet)
	nop
	nop
	nop
	mov     %g5,%fp                 ! restore sp,fp after psr write
	mov     %g6,%sp

#ifdef CYGPKG_HAL_SPARC_LEON3	
/*__sparc_leon3_cache_flush:*/
#define ASI_LEON3_IFLUSH    0x11
#define ASI_LEON3_DFLUSH    0x11
	sta %g0, [%g0] ASI_LEON3_IFLUSH
	sta %g0, [%g0] ASI_LEON3_DFLUSH
#else
/*__sparc_leon2_cache_flush:*/
#define ASI_LEON2_IFLUSH		0x05
#define ASI_LEON2_DFLUSH		0x06
	sta %g0, [%g0] ASI_LEON2_IFLUSH
	sta %g0, [%g0] ASI_LEON2_DFLUSH 
#endif
	
#ifndef CYGHWR_HAL_SPARC_MULTIPLE_VECTOR_TRAPPING
        set     1<<13,%g1       ! Single vector trapping, Leon3: asr17.bit13==1
        rd      %asr17,%g2
        or      %g1,%g2,%g2
	wr	%g2,%asr17	
        nop 
	nop
	nop
        .global single_vector_trapping_not_active
single_vector_trapping_not_active:      
        rd      %asr17,%g2                          !you have to enable Single Vector Trapping in the vhdl model
        andcc   %g2,%g1,%g2                         !you have to enable Single Vector Trapping in the vhdl model
         be     single_vector_trapping_not_active   !you have to enable Single Vector Trapping in the vhdl model
        nop
#endif // !CYGHWR_HAL_SPARC_MULTIPLE_VECTOR_TRAPPING

// INCLUDE PLATFORM BOOT

#ifdef CYGPKG_HAL_SMP_SUPPORT 
	 
	!smp  dbg
	!set 0x800000, %g2 
	lda [%g0] 2, %g1          ! cache cfg
	set 0x800000, %g2         ! snoop enable
        or %g2,%g1,%g2
        
        !set	0x800000, %g2     ! snooping,disable cache for test 
        
        sta %g2, [%g0] 2          ! cache cfg
	
	rd     %asr17,%g2
	srl    %g2,28,%g2
	
	! trap all cpu's except cpu nr 0
haltsmp:	
	cmp %g2,%g0
	 beq mastergo
	nop

        
	! slaves cpus will idle until master cpu (0) will issue a
	! mctrl->mpstatus write to wake them up. They will here branch to 
        ! cyg_hal_smp_start then
	
cyg_hal_smp_start:	
	! master called
	
	sethi	%hi(trap_table), %g1
	wr	%g1, %tbr	! Traps are at reset_vector
	nop
	nop
	nop	
	
	hal_smp_cpu %fp
	sll %fp,2,%fp
	set __interrupt_stack_vector,%i0
	ld [%i0+%fp], %fp
	
	sub	%fp, 16 * 4, %sp	! Stack pointer

#if (__WINSIZE == 2)
	wr	%g0, 0, %wim	! 2 windows 
#else        
	wr	%g0, __WIM_INIT, %wim	! Window 7 (current) is invalid
#endif                
        
	nop
	nop
	nop
	wr	%g0, 0xfe0 + __WIN_INIT, %psr
	nop				! Enable traps:             
	nop				! set psr, _do_ mask interrupts
	nop				! mode = prevMode = S, CWP=7

	
	call	cyg_hal_smp_startup
	 nop
	
		
mastergo:	
		
#endif
	
// This should set up RAM and caches, and calm down any external interrupt
// sources.  Also copy two instructions from real_vector_instructions
// into reset_vector, then invalidate the instruction cache.
	
#include <cyg/hal/halboot.si>
	
// halboot.si returns with %sp all set, in sleb versions.
// (though we override if there is an interrupt stack)

	! now set up a stack and initial frame linkage
	! so as to be able to make C function calls:
	! current window is 7, the highest, so we store a
	! saved frame thingy that refers to itself in the stack,
	! then another which is valid and drop into main from there.

	set cyg_hal_stack_top,%i0
        st %fp,[%i0]
        
	set cyg_interrupt_stack,%fp
#ifdef CYGPKG_HAL_SMP_SUPPORT
	set __interrupt_stack_vector,%i0
	st  %fp,[%i0]
#endif	
	! Clear Current Stack manually, it is 16-byte aligned
	! 
	set	cyg_interrupt_stack_base, %l0
	mov	%fp, %l1
clr_current_stack:
	std	%g0, [%l0]
	add	%l0, 16, %l0
	cmp	%l1, %l0
	bne	clr_current_stack
	std	%g0, [%l0-8]
	
	andn	%i6, 7, %i6		! round fp down to double alignment
	mov	0, %i7			! null return address
	sethi	%hi(0xb51ac000), %i0	! "BStac" pattern
	or	%i0, 24, %i0	
	or	%i0, 1, %i1
	or	%i0, 2, %i2
	or	%i0, 3, %i3
	or	%i0, 4, %i4
	or	%i0, 5, %i5

	sethi	%hi(0xb51ac000), %l0	! "BStac" pattern
	or	%l0, 16, %l0	
	or	%l0, 1, %l1
	or	%l0, 2, %l2
	or	%l0, 3, %l3
	or	%l0, 4, %l4
	or	%l0, 5, %l5
	or	%l0, 6, %l6
	or	%l0, 7, %l7
	
	sub	%fp, 16 * 4, %sp	! Stack pointer

	std	%l0, [%sp + 0 * 4]	! save L & I registers
	std	%l2, [%sp + 2 * 4]	! into new stack frame
	std	%l4, [%sp + 4 * 4]
	std	%l6, [%sp + 6 * 4]

	std	%i0, [%sp + 8 * 4]
	std	%i2, [%sp + 10 * 4]
	std	%i4, [%sp + 12 * 4]
	std	%i6, [%sp + 14 * 4]
	
	sethi	%hi(0xb0010000), %o0	! "Boot" pattern
	or	%o0, 8, %o0	
	or	%o0, 1, %o1
	or	%o0, 2, %o2
	or	%o0, 3, %o3
	or	%o0, 4, %o4
	or	%o0, 5, %o5

#if (__WINSIZE == 2)
	wr	%g0, 0, %wim	! 2 windows 
#else        
	wr	%g0, __WIM_INIT, %wim	! Window 7 (current) is invalid
#endif                
	nop
	nop
	nop
	
	sethi	%hi(0xb0010000), %g1	! "Boot" pattern
	or	%g1, 2, %g2
	or	%g1, 3, %g3
	or	%g1, 4, %g4
	or	%g1, 5, %g5
	or	%g1, 6, %g6
	or	%g1, 7, %g7
	or	%g1, 1, %g1

#ifdef CYGPKG_HAL_SPARC_LAZY_FPU_SWITCH
	set	0x0fe0 + __WIN_INIT, %g2	
#else	
	set	0x1fe0 + __WIN_INIT, %g2
#endif	
	wr	%g0, %g2, %psr
	or	%g1, 2, %g2	
	!wr	%g0, 0x1fe0 + __WIN_INIT, %psr
	nop				! Enable traps:             
	nop				! set psr, _do_ mask interrupts
	nop				! mode = prevMode = S, CWP=7
	
	
#if defined(CYGHWR_HAL_SPARC_FLAT)
       	save	%sp, -24 * 4, %sp       ! get valid window
#endif

	! now we can start calling out and running C code!
	.extern cyg_hal_start	
	call	cyg_hal_start		! puts return address in %o7
	or	%g1, 1, %g1
	
#ifdef CYGPKG_HAL_SMP_SUPPORT
	.extern cyg_hal_smp_init
	call cyg_hal_smp_init
	 nop
#endif
		
	call cyg_start /* does not return */
	 nop
	
	
loop_forever:
	ta	1
	b	loop_forever		! if it returns
	nop

#define SYM(n) n
#define PUBLIC(n) .globl SYM(n)
        
        /*
         *  system call
         *
         *  On entry:
         *
         *    l0 = psr (from trap table)
         *    l1 = pc
         *    l2 = npc
         *    i0 = system call id
         */

        PUBLIC(syscall)

SYM(syscall):

	subcc   %i0, 2, %g0		! syscall 2, disable interrupts
	bne 	3f
	or    	%l0, 0x0f00, %l4	! set PIL=15
	mov   	%l4, %psr
        or	%l0, SPARC_PSR_ET_MASK, %i0	! return old psr with ET=1
	ba,a	9f
3:
	subcc   %i0, 3, %g0		! syscall 3, enable interrupts
	bne 	1f
	and	%i1, SPARC_PSR_PIL_MASK, %l4
	andn	%l0, SPARC_PSR_PIL_MASK, %l5
	or	%l5, %l4, %l4
	mov   	%l4, %psr
        ba,a	9f

1:					
	ta	0			! halt 
9:					! leave
        jmpl    %l2, %g0
        rett    %l2 + 4


	
!---------------------------------------------------------------------------
! hal_vsr_table...

	.section ".bss.noclr"

	.balign 4
        .global cyg_hal_stack_top
cyg_hal_stack_top:
        .long 0

	.section ".data"        
	.balign 4
	.global hal_vsr_table
hal_vsr_table:
	.rept 16
	.word hal_default_interrupt_vsr
	.endr
	.rept 11
	.word hal_default_exception_vsr
	.endr

!---------------------------------------------------------------------------
! Bootup stack (only needed explicitly in sim)
	
#if defined(BOOTUPSTACK_IS_INTERRUPTSTACK) || \
    defined(CYGIMP_HAL_COMMON_INTERRUPTS_USE_INTERRUPT_STACK)
       
#ifndef CYGPKG_HAL_SMP_SUPPORT	
	
#ifndef CYGNUM_HAL_COMMON_INTERRUPTS_STACK_SIZE
#define CYGNUM_HAL_COMMON_INTERRUPTS_STACK_SIZE 4096
#endif
        
	.section ".bss.noclr"
	.balign 16
	.global cyg_interrupt_stack_base
cyg_interrupt_stack_base:
	.rept CYGNUM_HAL_COMMON_INTERRUPTS_STACK_SIZE
	.byte 0
	.endr
	.balign 16
	.global cyg_interrupt_stack
cyg_interrupt_stack:
	.long	0,0,0,0,0,0,0,0		! here be secret state stored
	
	.section ".bss"

#else
	.section ".bss.noclr"
	.global __interrupt_stack_vector
__interrupt_stack_vector:
	.rept CYGPKG_HAL_SMP_CPU_MAX
	.long 0
	.endr
	
	.section ".bss.noclr"
	.balign 16
	.global cyg_interrupt_stack_base, __interrupt_stack_first
cyg_interrupt_stack_base:
__interrupt_stack_base:
__interrupt_stack_first:
	.rept CYGNUM_HAL_COMMON_INTERRUPTS_STACK_SIZE
	.byte 0
	.endr
	.balign 16
	.global cyg_interrupt_stack
cyg_interrupt_stack:
	
	.section ".bss"
__interrupt_stack:
	.rept CYGNUM_HAL_COMMON_INTERRUPTS_STACK_SIZE*(CYGPKG_HAL_SMP_CPU_MAX-1)
	.byte 0
	.endr
	.long	0,0,0,0,0,0,0,0		! here be secret state stored
	
#endif
	
#endif

!------------------------------------------------------------------------
! Define a section that reserves space at the start of RAM for the
! vectors to be copied into, for ROM start only.
	
#ifdef CYGIMP_HAL_SPARC_COPY_VECTORS_TO_RAM
	.section ".ram_vectors","awx",@nobits
	! need a space at base of RAM for copied vector/trampoline code
	.align 0x1000
	.space 8			! for fencepost errors
	.space (rom_vectors_end - rom_vectors)
#endif // CYGIMP_HAL_SPARC_COPY_VECTORS_TO_RAM

!------------------------------------------------------------------------
! end of vectors.S
