//===========================================================================
//
//      hal_amba.c
//
//      SPARC Architecture LEON3-specific private variables
//
//===========================================================================
//####ECOSGPLCOPYRIGHTBEGIN####
// -------------------------------------------
// This file is part of eCos, the Embedded Configurable Operating System.
// Copyright (C) 1998, 1999, 2000, 2001, 2002 Red Hat, Inc.
//
// eCos is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later version.
//
// eCos is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with eCos; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
//
// As a special exception, if other files instantiate templates or use macros
// or inline functions from this file, or you compile this file and link it
// with other works to produce a work based on this file, this file does not
// by itself cause the resulting work to be covered by the GNU General Public
// License. However the source code for this file must still be made available
// in accordance with section (3) of the GNU General Public License.
//
// This exception does not invalidate any other reasons why a work based on
// this file might be covered by the GNU General Public License.
//
// Alternative licenses for eCos may be arranged by contacting Red Hat, Inc.
// at http://sources.redhat.com/ecos/ecos-license/
// -------------------------------------------
//####ECOSGPLCOPYRIGHTEND####
//===========================================================================
//#####DESCRIPTIONBEGIN####
//
// Author(s):    Gaisler Research (Konrad Eisele<eiselekd@web.de>)
// Contributors: 
// Date:         2004
// Purpose:      private amba for LEON3 processor.
//              
//####DESCRIPTIONEND####
//
//===========================================================================

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <cyg/hal/hal_leon3.h>
#include <cyg/infra/diag.h>

//#define DEBUG_CONFIG

/* Structure containing address to devices found on the Amba Plug&Play bus */
amba_confarea_type amba_conf;

/* Pointers to Interrupt Controller configuration registers */
volatile LEON3_IrqCtrl_Regs_Map *LEON3_IrqCtrl_Regs __attribute__((section(".bss.noclr")));
volatile LEON3_GpTimer_Regs_Map *LEON3_GpTimer_Regs = 0;
volatile LEON3_APBUART_Regs_Map *LEON3_APBUART_Regs __attribute__((section(".bss.noclr")));
unsigned long LEON3_GpTimer_Irq = 0;

static void vendor_dev_string(unsigned long conf, char *vendorbuf,char *devbuf) {
  int vendor = amba_vendor(conf); int dev = amba_device(conf);
  char *devstr; char *vendorstr; 
  sprintf(vendorbuf, "Unknown vendor %2x",vendor);
  sprintf(devbuf, "Unknown device %2x",dev);
  vendorstr = vendor_id2str(vendor);
  if (vendorstr) {
    sprintf(vendorbuf, "%s",vendorstr);
  } 
  devstr = device_id2str(vendor,dev);
  if (devstr) {
    sprintf(devbuf, "%s",devstr);
  }
}

void amba_prinf_config(void)  
{
  char devbuf[128]; char vendorbuf[128]; unsigned int conf;
  int i = 0; int j = 0; 
  unsigned int addr; unsigned int m;
  diag_printf("             Vendors         Slaves\n");
  diag_printf("Ahb masters:\n");
  i = 0;
  while (i < amba_conf.ahbmst.devnr) 
  {
    conf = amba_get_confword(amba_conf.ahbmst, i, 0);
    vendor_dev_string(conf,vendorbuf,devbuf);
    diag_printf("%2i(%2x:%3x|%2i): %16s %16s \n", i, amba_vendor(conf), amba_device(conf), amba_irq(conf), vendorbuf, devbuf);
    for (j = 0;j < 4;j++) {
      m = amba_ahb_get_membar(amba_conf.ahbmst,i,j);
      if (m) {
        addr = amba_membar_start(m);
        diag_printf(" +%i: 0x%x \n", j, addr);
      }
    }
    i++;
  }
  diag_printf("Ahb slaves:\n");
  i = 0;
  while (i < amba_conf.ahbslv.devnr) 
  {
    conf = amba_get_confword(amba_conf.ahbslv, i, 0);
    vendor_dev_string(conf,vendorbuf,devbuf);
    diag_printf("%2i(%2x:%3x|%2i): %16s %16s \n", i, amba_vendor(conf), amba_device(conf), amba_irq(conf), vendorbuf, devbuf);
    for (j = 0;j < 4;j++) {
      m = amba_ahb_get_membar(amba_conf.ahbslv,i,j);
      if (m) {
        addr = amba_membar_start(m);
	if (amba_membar_type(m) == AMBA_TYPE_AHBIO) {
	  addr = AMBA_TYPE_AHBIO_ADDR(addr);
	} else if (amba_membar_type(m) == AMBA_TYPE_APBIO) {
	  diag_printf("Warning: apbio membar\n");
	}
	diag_printf(" +%i: 0x%x (raw:0x%x)\n", j, addr, m);
      }
    }
    i++;
  }
  diag_printf("Apb slaves:\n");
  i = 0;
  while (i < amba_conf.apbslv.devnr) 
  {
    
    conf = amba_get_confword(amba_conf.apbslv, i, 0);
    vendor_dev_string(conf,vendorbuf,devbuf);
    diag_printf("%2i(%2x:%3x|%2i): %16s %16s \n", i, amba_vendor(conf), amba_device(conf), amba_irq(conf), vendorbuf, devbuf);
    
    m = amba_apb_get_membar(amba_conf.apbslv, i);
    addr = amba_iobar_start(amba_conf.apbmst, m);
    diag_printf(" +%2i: 0x%x (raw:0x%x) \n", 0, addr, m);
    
    i++;
    
  }
  
}

#define amba_insert_device(tab, address) \
{ \
  if (LEON3_BYPASS_LOAD_PA(address)) \
  { \
    (tab)->addr[(tab)->devnr] = (address); \
    (tab)->devnr ++; \
  } \
} while(0)

/*
 *  Used to scan system bus. Probes for AHB masters, AHB slaves and 
 *  APB slaves. Addresses to configuration areas of the AHB masters,
 *  AHB slaves, APB slaves and APB master are storeds in 
 *  amba_ahb_masters, amba_ahb_slaves and amba.
 */

void amba_init(void) 
{
  unsigned int *cfg_area;  /* address to configuration area */
  unsigned int mbar, conf;
  int i, j;
  

  memset(&amba_conf,0,sizeof(amba_conf));
  //amba_conf.ahbmst.devnr = 0; amba_conf.ahbslv.devnr = 0; amba_conf.apbslv.devnr = 0;
  
  cfg_area = (unsigned int *) (LEON3_IO_AREA | LEON3_CONF_AREA);

  for (i = 0; i < LEON3_AHB_MASTERS; i++) 
  {
    amba_insert_device(&amba_conf.ahbmst, cfg_area);
    cfg_area += LEON3_AHB_CONF_WORDS;
  }

  cfg_area = (unsigned int *) (LEON3_IO_AREA | LEON3_CONF_AREA | LEON3_AHB_SLAVE_CONF_AREA);
  for (i = 0; i < LEON3_AHB_SLAVES; i++) 
  {
    amba_insert_device(&amba_conf.ahbslv, cfg_area);
    cfg_area += LEON3_AHB_CONF_WORDS;
  }  

  for (i = 0; i < amba_conf.ahbslv.devnr; i ++) 
  {
    conf = amba_get_confword(amba_conf.ahbslv, i, 0);
    mbar = amba_ahb_get_membar(amba_conf.ahbslv, i, 0);
    if ((amba_vendor(conf) == VENDOR_GAISLER) && (amba_device(conf) == GAISLER_APBMST))
    {
      amba_conf.apbmst = amba_membar_start(mbar);
      cfg_area = (unsigned int *) (amba_conf.apbmst | LEON3_CONF_AREA);
      
      for (j = amba_conf.apbslv.devnr; j < LEON3_APB_SLAVES; j++)
      {
	amba_insert_device(&amba_conf.apbslv, cfg_area);
	cfg_area += LEON3_APB_CONF_WORDS;
      }
      break;
    }
  }    
 
  /* Find LEON3 Interrupt controler */
  LEON3_IrqCtrl_Regs = (volatile LEON3_IrqCtrl_Regs_Map *) amba_find_apbslv_addr(VENDOR_GAISLER, GAISLER_IRQMP, 0);
  LEON3_GpTimer_Regs = (volatile LEON3_GpTimer_Regs_Map *) amba_find_apbslv_addr(VENDOR_GAISLER, GAISLER_GPTIMER, &LEON3_GpTimer_Irq);
  LEON3_APBUART_Regs = (volatile LEON3_APBUART_Regs_Map *) amba_find_apbslv_addr(VENDOR_GAISLER, GAISLER_APBUART, 0);
}

unsigned long amba_find_apbslv_addr(unsigned long vendor, unsigned long device, unsigned long *irq) {
  unsigned int i,conf,iobar;
  for (i = 0; i < amba_conf.apbslv.devnr; i++) {
    conf = amba_get_confword(amba_conf.apbslv, i, 0);
    if ((amba_vendor(conf) == vendor) && (amba_device(conf) == device))
    {
      if (irq) {
        *irq = amba_irq(conf);
      }
      iobar = amba_apb_get_membar(amba_conf.apbslv, i);
      return amba_iobar_start(amba_conf.apbmst, iobar);
    }
  }
  return 0;
}

char *gaisler_device_str(int id) {
  switch(id) {
  case GAISLER_LEON3:    return "GAISLER_LEON3";
  case GAISLER_LEON3DSU: return "GAISLER_LEON3DSU";
  case GAISLER_ETHAHB:   return "GAISLER_ETHAHB";
  case GAISLER_ETHMAC:   return "GAISLER_ETHMAC";
  case GAISLER_APBMST:   return "GAISLER_APBMST";
  case GAISLER_AHBUART:  return "GAISLER_AHBUART";
  case GAISLER_SRCTRL:   return "GAISLER_SRCTRL";  
  case GAISLER_SDCTRL:   return "GAISLER_SDCTRL"; 
  case GAISLER_APBUART:  return "GAISLER_APBUART"; 
  case GAISLER_IRQMP:    return "GAISLER_IRQMP"; 
  case GAISLER_AHBRAM:   return "GAISLER_AHBRAM"; 
  case GAISLER_GPTIMER:  return "GAISLER_GPTIMER";
  case GAISLER_PCITRG:   return "GAISLER_PCITRG"; 
  case GAISLER_PCISBRG:  return "GAISLER_PCISBRG"; 
  case GAISLER_PCIFBRG:  return "GAISLER_PCIFBRG"; 
  case GAISLER_PCITRACE: return "GAISLER_PCITRACE"; 
  case GAISLER_AHBTRACE: return "GAISLER_AHBTRACE";
  case GAISLER_ETHDSU:   return "GAISLER_ETHDSU";
  case GAISLER_PIOPORT:  return "GAISLER_PIOPORT";
 
  case GAISLER_L2TIME:   return "GAISLER_L2TIME";
  case GAISLER_L2C:      return "GAISLER_L2C";
  case GAISLER_PLUGPLAY: return "GAISLER_PLUGPLAY";
    
  default: break;
  }
  return 0;
}

char *esa_device_str(int id) {
  switch(id) {
  case ESA_LEON2:  return "ESA_LEON2";
  case ESA_MCTRL:  return "ESA_MCTRL";
  default: break;
  }
  return 0;
}

char *opencores_device_str(int id) {
  switch(id) {
  case OPENCORES_PCIBR:  return "OPENCORES_PCIBR";
  case OPENCORES_ETHMAC:  return "OPENCORES_ETHMAC";
  default: break;
  }
  return 0;
}

char *ntnu_device_str(int id) {
  switch(id) {
  case NTNU_TMU:  return "NTNU_TMU";
  default: break;
  }
  return 0;
}

char *device_id2str(int vendor, int id) {
  switch(vendor) {
  case VENDOR_GAISLER:    return gaisler_device_str(id);
  case VENDOR_ESA:        return esa_device_str(id);
  case VENDOR_OPENCORES:  return opencores_device_str(id);
  case VENDOR_NTNU:       return ntnu_device_str(id);
  case VENDOR_PENDER:
  default: break;
  }
  return 0;
}

char *vendor_id2str(int vendor) {
  switch(vendor) {
  case VENDOR_GAISLER:    return "VENDOR_GAISLER";
  case VENDOR_ESA:        return "VENDOR_ESA";
  case VENDOR_OPENCORES:  return "VENDOR_OPENCORES";
  case VENDOR_NTNU:       return "VENDOR_NTNU";
  case VENDOR_PENDER:     return "VENDOR_PENDER";
  default: break;
  }
  return 0;
}
