//===========================================================================
//
//      hal_irq.c
//
//      SPARC Architecture LEON3-specific private variables
//
//===========================================================================
//####ECOSGPLCOPYRIGHTBEGIN####
// -------------------------------------------
// This file is part of eCos, the Embedded Configurable Operating System.
// Copyright (C) 1998, 1999, 2000, 2001, 2002 Red Hat, Inc.
//
// eCos is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later version.
//
// eCos is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with eCos; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
//
// As a special exception, if other files instantiate templates or use macros
// or inline functions from this file, or you compile this file and link it
// with other works to produce a work based on this file, this file does not
// by itself cause the resulting work to be covered by the GNU General Public
// License. However the source code for this file must still be made available
// in accordance with section (3) of the GNU General Public License.
//
// This exception does not invalidate any other reasons why a work based on
// this file might be covered by the GNU General Public License.
//
// Alternative licenses for eCos may be arranged by contacting Red Hat, Inc.
// at http://sources.redhat.com/ecos/ecos-license/
// -------------------------------------------
//####ECOSGPLCOPYRIGHTEND####
//===========================================================================
//#####DESCRIPTIONBEGIN####
//
// Author(s):    Gaisler Research (Konrad Eisele<eiselekd@web.de>)
// Contributors: 
// Date:         2004
// Purpose:      private amba for LEON3 processor.
//              
//####DESCRIPTIONEND####
//
//===========================================================================

#ifdef CYGPKG_KERNEL
#include <pkgconf/kernel.h>
#endif

#include <cyg/infra/cyg_type.h>         // Base types
#include <cyg/infra/cyg_trac.h>         // tracing macros
#include <cyg/infra/cyg_ass.h>          // assertion macros
#include <cyg/infra/diag.h>

#include <cyg/hal/hal_intr.h>
#include <cyg/hal/hal_io.h>
#include <cyg/hal/hal_leon3.h>
#include <cyg/hal/hal_smp.h>

#define LEON_HARD_INT(x) (1 << (x)) 
externC volatile LEON3_IrqCtrl_Regs_Map *LEON3_IrqCtrl_Regs; 
#define LEON_IMASK(i) ((&LEON3_IrqCtrl_Regs ->mask[i]))
#define LEON_IFORCE(i) ((&LEON3_IrqCtrl_Regs ->force[i]))

#define get_irqmask(irq) LEON_HARD_INT(irq)

/*static inline unsigned long get_irqmask(unsigned int irq)
{
	unsigned long mask;
	if (!irq || irq > 0xf) {
	  //diag_printf("leon_get_irqmask: false irq number\n");
	  mask = 0;
	} else {
	  mask = LEON_HARD_INT(irq);
	}
	return mask;
}*/

#ifndef CYGPKG_HAL_SMP_SUPPORT

void leon3_disable_irq(unsigned int irq_nr)
{
	unsigned long mask, v;
	mask = get_irqmask(irq_nr) & LEON_IRQMASK_R;
	HAL_READ_UINT32(LEON_IMASK(0),v);
	HAL_WRITE_UINT32(LEON_IMASK(0), (v & ~(mask)));
	
}

void leon3_enable_irq(unsigned int irq_nr)
{
	unsigned long mask,v;
	mask = get_irqmask(irq_nr) & LEON_IRQMASK_R;
	HAL_READ_UINT32(LEON_IMASK(0),v);
	HAL_WRITE_UINT32(LEON_IMASK(0), (v | (mask)));
}

#else

unsigned long irq_mask;
unsigned long cpu_route_mask[CYGPKG_HAL_SMP_CPU_MAX];
HAL_SPINLOCK_TYPE cyg_hal_smp_smpirq_lock = HAL_SPINLOCK_INIT_CLEAR;

/* route all irq to 0 */
void leon3_smpirq_init(void) {
  int i;
  /*
  for (i=0; i<CYGPKG_HAL_SMP_CPU_MAX; i++)
    cpu_route_mask[i] = -1;
  */
  cpu_route_mask[0] = -1;
}

void leon3_disable_irq(unsigned int irq_nr)
{
  unsigned long mask, cpu;

  //diag_printf("disable %d\n",irq_nr);
  
  mask = get_irqmask(irq_nr) & LEON_IRQMASK_R;
  irq_mask &= ~mask;
  HAL_SPINLOCK_SPIN( cyg_hal_smp_smpirq_lock );
  for (cpu = 0; cpu < HAL_SMP_CPU_COUNT(); cpu++) {
    mask = cpu_route_mask[cpu] & irq_mask;
    HAL_WRITE_UINT32(LEON_IMASK(cpu), mask);
  }
  HAL_SPINLOCK_CLEAR( cyg_hal_smp_smpirq_lock );
}

void leon3_enable_irq(unsigned int irq_nr)
{
  unsigned long mask, cpu, v = 0;
  mask = get_irqmask(irq_nr) & LEON_IRQMASK_R;

  //diag_printf("enable %d\n",irq_nr);

  
  for (cpu = 0; cpu < HAL_SMP_CPU_MAX; cpu++) {
    v |= cpu_route_mask[cpu] & mask;
  }
  if (!v) {
    cpu_route_mask[0] |= mask;
  }

  irq_mask |= mask;
  HAL_SPINLOCK_SPIN( cyg_hal_smp_smpirq_lock );
  for (cpu = 0; cpu < HAL_SMP_CPU_MAX; cpu++) {
    mask = (cpu_route_mask[cpu] | 0x04000) & irq_mask;
    //mask = cpu_route_mask[cpu] & irq_mask;
    //mask = mask | 0x04000;
    HAL_WRITE_UINT32(LEON_IMASK(cpu), mask);
  }
  HAL_SPINLOCK_CLEAR( cyg_hal_smp_smpirq_lock );
}

void leon3_interrupt_set_cpu( CYG_WORD32 vector, HAL_SMP_CPU_TYPE cpu ) {
  unsigned long mask, i;

  HAL_SPINLOCK_SPIN( cyg_hal_smp_smpirq_lock );
  mask = get_irqmask(vector) & LEON_IRQMASK_R;
  for (i = 0; i < HAL_SMP_CPU_COUNT(); i++) {
    if ( i == cpu) {
      cpu_route_mask[i] |= mask;
    } else {
      cpu_route_mask[i] &= ~mask;
    }
    //mask = cpu_route_mask[i] & irq_mask;
    HAL_WRITE_UINT32(LEON_IMASK(i), cpu_route_mask[i] & irq_mask);
  }
  HAL_SPINLOCK_CLEAR( cyg_hal_smp_smpirq_lock );
}

void leon3_interrupt_get_cpu( CYG_WORD32 vector, HAL_SMP_CPU_TYPE *cpu ){
  unsigned long mask, i;

  mask = get_irqmask(vector) & LEON_IRQMASK_R;
  *cpu = 0;
  for (i = 0; i < HAL_SMP_CPU_COUNT(); i++) {
    if (cpu_route_mask[i] & mask) {
      *cpu = i;
    }
  }
}


void leon3_smpirq_raise(CYG_WORD32 cpu) {
    unsigned long mask = get_irqmask(CYGNUM_HAL_SMP_CPU_INTERRUPT_VECTOR(cpu)) & LEON_IRQMASK_R;
    unsigned long  i, v;
    unsigned long masksaved[HAL_SMP_CPU_MAX];
    HAL_SMP_CPU_TYPE me = HAL_SMP_CPU_THIS();
    CYG_INTERRUPT_STATE istate = 0;    
    
    HAL_DISABLE_INTERRUPTS( istate );

    /*if (cpu == me) {
      while(1);
      }*/
    
    HAL_SPINLOCK_SPIN( cyg_hal_smp_smpirq_lock );
    
    HAL_WRITE_UINT32(LEON_IFORCE(cpu), (mask));
    
    
    /* for (i = 0; i < HAL_SMP_CPU_COUNT(); i++) { */
/*       HAL_READ_UINT32(LEON_IMASK(i),v); */
/*       masksaved[i] = v; */
/*       HAL_WRITE_UINT32(LEON_IMASK(i), (v & ~(mask))); */
/*     } */
    
/*     HAL_READ_UINT32(LEON_IMASK(cpu),v); */
/*     HAL_WRITE_UINT32(LEON_IMASK(cpu), (v | (mask))); */
/*     HAL_READ_UINT32(LEON_IFORCE(cpu),v); */
/*     HAL_WRITE_UINT32(LEON_IFORCE(cpu), (v | (mask))); */

/*     /\*do { */
/*       HAL_READ_UINT32(LEON_IFORCE(cpu),force); */
/*       } while( (force & mask) != 0 ); *\/ */

/*     for (i = 0; i < HAL_SMP_CPU_COUNT(); i++) { */
/*       HAL_WRITE_UINT32(LEON_IMASK(i), (masksaved[i])); */
/*     } */
    
    HAL_SPINLOCK_CLEAR( cyg_hal_smp_smpirq_lock );
    
    HAL_RESTORE_INTERRUPTS( istate );
}
    
#endif
