// file:  tmu_test.c
// descr: test thu TMU functionality
// notes:
// file must be (or have a symbolic link) in the same directory as
// the leon3 design. The global software makefile grlib/software/leon3/Makefile
// must have the PROGS variable extended with "tmu_test" and the systest.c file
// in the design directory must do the actual function call:
//   tmu_test(addr, irq);

#include "testmod.h"  // extern fail()
#include "irqmp.h"    // irqmp_base

#ifndef TMUADDR
#define TMUADDR 0x80001000
#endif
#define MAXTIMERS 32    // tar dette plass i minnet??
#define TMRTASK 0
#define TMRINT1 1
#define TMRINT2 2
#define TMRINT3 3

struct timer_regs {
  volatile unsigned int count;	        // 0x0
  volatile unsigned int limit;		// 0x4
  volatile unsigned int period;		// 0x8
  volatile unsigned int control;	// 0xC bit0:irqen, bit1:irqpen
  //  volatile unsigned int dummy;	        // 0xC
};

struct tmu {
  volatile unsigned int control;        // 0x00 b0:enable tmu, b1:enable irq, 2-6 ntimers, 7-12 timerbits
  volatile unsigned int tsel;		// 0x04 timer select reg
  volatile unsigned int irqpen;		// 0x08 irqpen for all timers
  volatile unsigned int dummy;		// 0x0C unused
  //volatile unsigned int irqen;        // 0x0C interrupt enable for all tmrs
  struct timer_regs timer[MAXTIMERS];
};

static volatile tmuirq = 0;
static struct tmu *p = (struct tmu *)TMUADDR; // base pointer to TMU regs

inline
void tmu_enable(void)
{
  p->control = 0x1; // set global tmu enable bit
}


inline
void tmu_disable(void)
{
  p->control = 0x0;
}

inline
int tmu_ntimers(void)
{
  return (p->control >> 1) & 0x1f;
}

inline
int tmu_tbits(void)
{
  return (p->control >> 6) & 0x3f;
}


inline
void tmu_select_tmr(int timer)
{
  p->tsel = timer;
}


inline
int tmu_cur_tmr(void)
{
  return p->tsel;
}


inline
void tmu_enable_irq(int timer)
{
  p->timer[timer].control |= 0x1; // set irqen bit
  return;
}

inline
void tmu_disable_irq(int timer)
{
  p->timer[timer].control &= (~0x1); // clear irqen bit
  return;
}


inline
void tmu_clear_irqpen(void)
{
  p->irqpen = 0;
  return;
}


inline
void tmu_tmr_setup(int timer, int limit, int period, int irqen)
{
  //  p->tsel = timer;
  //  p->timer[timer].count = 0;
  p->timer[timer].limit = limit;
  p->timer[timer].period = period;
  if (irqen != 0) {
    tmu_enable_irq(timer);
  } else {
    tmu_disable_irq(timer);
  }

  return;
}


static tmu_irqhandler(int irq)
{
  //  printf("enter isr\n");
  tmuirq += 1;
  tmu_clear_irqpen();
  //printf("exit isr\n");
}


void tmu_test(int addr, int irq)
{
  p = (struct tmu *)addr; // base pointer to TMU regs

  int i;
  const int t = 0; // task timer
  int limit = 300;
  int period = 500;
  int ntimers = 0;



  tmu_select_tmr(t);
  if (tmu_cur_tmr() != t) {
    fail(1);
  }
  //  p->tsel = t;
  //if (p->tsel != t) {
  //  fail(1);
  //}

  // check all timers for read/write
  tmu_disable();
  //p->control = 0x0;
  ntimers = tmu_ntimers();
  for (i=0; i<ntimers; i++) {
    tmu_tmr_setup(i, limit, period, 0);
    //p->timer[i].limit = limit;
    //p->timer[i].period = period;
    if (p->timer[i].limit != limit) {
      fail(31);
    }
    if (p->timer[i].period != period) {
      fail(32);
    }
  }

  if (p->irqpen != 0)
    fail(40); // interrupt should not have been enabled yet!
    //printf("irqpen is not 0!\n");


  // irqtest() must be run prior to this (to set up irqmp_base)
  static struct irqmp *irqmp_base = (struct irqmp *)0x80000200;

  if (irqmp_base) {
    catch_interrupt(tmu_irqhandler, irq);
    init_irqmp(irqmp_base);
    irqmp_base->irqmask = (1 << irq);	  // unmask interrupt
    tmu_enable_irq(0);  // enable irq for tmr0
    tmu_enable();
    while (tmuirq < 2) {
      //      printf("waiting on tmu interrupts\n");
    }
    //    asm("wr %g0, %g0, %asr19");		// power-down
  }

  return;
}
